/* This file is part of SLB
   Copyright (C) 2011 Sergey Poznyakoff

   SLB is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   SLB is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with SLB.  If not, see <http://www.gnu.org/licenses/>. */

#include "slb.h"

int log_to_stderr = -1; /* -1 means autodetect */
int syslog_facility = LOG_DAEMON;
const char *syslog_tag = NULL;
int syslog_include_prio; /* syslog messages include priority */

struct log_trans {
	char *kw;
	int tok;
};

static struct log_trans facility_trans[] = {
	{ "USER",    LOG_USER },   
	{ "DAEMON",  LOG_DAEMON },
	{ "AUTH",    LOG_AUTH },
	{ "AUTHPRIV",LOG_AUTHPRIV },
	{ "MAIL",    LOG_MAIL },
	{ "CRON",    LOG_CRON },
	{ "LOCAL0",  LOG_LOCAL0 },
	{ "LOCAL1",  LOG_LOCAL1 },
	{ "LOCAL2",  LOG_LOCAL2 },
	{ "LOCAL3",  LOG_LOCAL3 },
	{ "LOCAL4",  LOG_LOCAL4 },
	{ "LOCAL5",  LOG_LOCAL5 },
	{ "LOCAL6",  LOG_LOCAL6 },
	{ "LOCAL7",  LOG_LOCAL7 },
	{ NULL }
};

static int
syslog_to_n(struct log_trans *trans, const char *str, int *pint)
{
	int i;

	for (i = 0; trans[i].kw; i++) {
		if (strcasecmp(trans[i].kw, str) == 0) {
			*pint = trans[i].tok;
			return 0;
		}
	}
	return 1;
}

static int
n_to_syslog(struct log_trans *trans, int tok, const char **pstr)
{
	int i;

	for (i = 0; trans[i].kw; i++) {
		if (trans[i].tok == tok) {
			*pstr = trans[i].kw;
			return 0;
		}
	}
	return 1;
}

static struct log_trans prio_trans[] = {
	{ "EMERG", LOG_EMERG },
	{ "ALERT", LOG_ALERT },
	{ "CRIT", LOG_CRIT },
	{ "ERR", LOG_ERR },
	{ "WARNING", LOG_WARNING },
	{ "NOTICE", LOG_NOTICE },
	{ "INFO", LOG_INFO },
	{ "DEBUG", LOG_DEBUG },
	{ NULL }
};

int
string_to_syslog_facility(const char *str, int *pfacility)
{
	return syslog_to_n(facility_trans, str, pfacility);
}

/* Logging */
void
syslog_printer(int prio, grecs_locus_t *locus, const char *fmt, va_list ap)
{
	static char *fmtbuf;
	static size_t fmtsize;
	size_t size;
	char linebuf[128];
	size_t locsize = 0;
	
	if (locus) {
		snprintf(linebuf, sizeof(linebuf), "%u", locus->line);
		locsize = strlen(locus->file) + 1 + strlen(linebuf) + 2;
	}
	if (syslog_include_prio) {
		const char *p;

		if (n_to_syslog(prio_trans, prio, &p) == 0) {
			size = strlen(p) + 3 + strlen(fmt) + locsize + 1;

			if (size > fmtsize) {
				fmtsize = size;
				fmtbuf = x2realloc(fmtbuf, &fmtsize);
			}
			if (locsize) {
				sprintf(fmtbuf, "%s:%s: [%s] %s",
					locus->file, linebuf, p, fmt);
				locsize = 0;
			} else
				sprintf(fmtbuf, "[%s] %s", p, fmt);
			fmt = fmtbuf;
		}
	}
	if (locsize) {
		size = strlen(fmt) + locsize + 1;
		if (size > fmtsize) {
			fmtsize = size;
			fmtbuf = x2realloc(fmtbuf, &fmtsize);
		}
		sprintf(fmtbuf, "%s:%s: %s", locus->file, linebuf, fmt);
	}
	vsyslog(prio, fmt, ap);
}

void
stderr_printer(int prio, grecs_locus_t *locus, const char *fmt, va_list ap)
{
	const char *p;
	fprintf (stderr, "%s: ", syslog_tag);

	if (locus)
		fprintf(stderr, "%s:%u: ", locus->file, locus->line);
	if (n_to_syslog(prio_trans, prio, &p) == 0)
		fprintf(stderr, "[%s] ", p);
	vfprintf(stderr, fmt, ap);
	fputc('\n', stderr);
}

static void (*log_printer)(int prio, grecs_locus_t *locus,
			   const char *fmt, va_list ap) = stderr_printer;

void
logger_setup()
{
	if (!syslog_tag) {
		syslog_tag = strrchr(program_name, '/');
		if (syslog_tag)
			syslog_tag++;
		else
			syslog_tag = program_name;
	}
	log_printer = log_to_stderr ? stderr_printer : syslog_printer;
	if (log_to_stderr)
		log_printer = stderr_printer;
	else {
		log_printer = syslog_printer;
		openlog(syslog_tag, LOG_PID, syslog_facility);
	}
	grecs_log_to_stderr = log_to_stderr;
}

void
vlogmsg(int prio, const char *fmt, va_list ap)
{
	log_printer(prio, NULL, fmt, ap);
}

void
logmsg(int prio, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	log_printer(prio, NULL, fmt, ap);
	va_end(ap);
}

void
logmsg_at_locus(int prio, grecs_locus_t *locus, const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	log_printer(prio, locus, fmt, ap);
	va_end(ap);
}

void
debug_printf(const char *fmt, ...)
{
	va_list ap;

	va_start(ap, fmt);
	log_printer(LOG_DEBUG, NULL, fmt, ap);
	va_end(ap);
}
